import { Map, OrderedSet } from "immutable";
import { Type, ClassType, EnumType, UnionType, TypeKind } from "./Type";
import { Namespace, Name, Namer, DependencyName } from "./Naming";
import { Renderer, BlankLineLocations } from "./Renderer";
import { StringMap } from "./Support";
import { Sourcelike } from "./Source";
import { Declaration } from "./DeclarationIR";
export declare abstract class ConvenienceRenderer extends Renderer {
    protected forbiddenWordsNamespace: Namespace;
    protected globalNamespace: Namespace;
    private _nameStoreView;
    private _propertyNamesStoreView;
    private _memberNamesStoreView;
    private _caseNamesStoreView;
    private _namedTypeNamer;
    private _classPropertyNamer;
    private _unionMemberNamer;
    private _enumCaseNamer;
    private _declarationIR;
    private _namedTypes;
    private _namedClasses;
    private _namedEnums;
    private _namedUnions;
    private _haveUnions;
    private _cycleBreakerTypes?;
    private _alphabetizeProperties;
    readonly topLevels: Map<string, Type>;
    protected readonly forbiddenNamesForGlobalNamespace: string[];
    protected forbiddenForClassProperties(_c: ClassType, _className: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected forbiddenForUnionMembers(_u: UnionType, _unionName: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected forbiddenForEnumCases(_e: EnumType, _enumName: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected topLevelDependencyNames(_t: Type, _topLevelName: Name): DependencyName[];
    protected namedTypeDependencyNames(_t: Type, _name: Name): DependencyName[];
    protected abstract topLevelNameStyle(rawName: string): string;
    protected abstract makeNamedTypeNamer(): Namer;
    protected abstract makeClassPropertyNamer(): Namer | null;
    protected abstract makeUnionMemberNamer(): Namer | null;
    protected abstract makeEnumCaseNamer(): Namer | null;
    protected abstract emitSourceStructure(givenOutputFilename: string): void;
    protected namedTypeToNameForTopLevel(type: Type): Type | undefined;
    protected readonly unionMembersInGlobalNamespace: boolean;
    protected readonly enumCasesInGlobalNamespace: boolean;
    protected readonly needsTypeDeclarationBeforeUse: boolean;
    protected canBeForwardDeclared(_t: Type): boolean;
    protected unionNeedsName(u: UnionType): boolean;
    protected setUpNaming(): Namespace[];
    private addDependenciesForNamedType;
    private nameForTopLevel;
    private addNamedForNamedType;
    private addPropertyNames;
    private makeUnionMemberName(u, unionName, t);
    private addUnionMemberNames;
    private addEnumCaseNames;
    private childrenOfType;
    protected readonly namedUnions: OrderedSet<UnionType>;
    protected readonly haveNamedUnions: boolean;
    protected readonly haveUnions: boolean;
    protected readonly enums: OrderedSet<EnumType>;
    protected readonly haveEnums: boolean;
    protected proposedUnionMemberNameForTypeKind: (_kind: TypeKind) => string | null;
    protected proposeUnionMemberName(_u: UnionType, _unionName: Name, fieldType: Type, lookup: (n: Name) => string): string;
    protected nameForNamedType: (t: Type) => Name;
    protected isForwardDeclaredType(t: Type): boolean;
    protected isImplicitCycleBreaker(_t: Type): boolean;
    protected canBreakCycles(_t: Type): boolean;
    protected isCycleBreakerType(t: Type): boolean;
    protected forEachTopLevel: (blankLocations: BlankLineLocations, f: (t: Type, name: Name) => void, predicate?: ((t: Type) => boolean) | undefined) => void;
    protected forEachDeclaration(blankLocations: BlankLineLocations, f: (decl: Declaration) => void): void;
    setAlphabetizeProperties: (value: boolean) => void;
    protected forEachClassProperty: (c: ClassType, blankLocations: BlankLineLocations, f: (name: Name, jsonName: string, t: Type) => void) => void;
    protected nameForUnionMember: (u: UnionType, t: Type) => Name;
    protected forEachUnionMember: (u: UnionType, members: OrderedSet<Type> | null, blankLocations: BlankLineLocations, sortOrder: ((n: Name, t: Type) => string) | null, f: (name: Name, t: Type) => void) => void;
    protected forEachEnumCase: (e: EnumType, blankLocations: BlankLineLocations, f: (name: Name, jsonName: string) => void) => void;
    protected callForNamedType<T extends Type>(t: T, f: (t: T, name: Name) => void): void;
    protected forEachSpecificNamedType<T extends Type>(blankLocations: BlankLineLocations, types: OrderedSet<T>, f: (t: T, name: Name) => void): void;
    protected forEachClass: (blankLocations: BlankLineLocations, f: (c: ClassType, className: Name) => void) => void;
    protected forEachEnum: (blankLocations: BlankLineLocations, f: (u: EnumType, enumName: Name) => void) => void;
    protected forEachUnion: (blankLocations: BlankLineLocations, f: (u: UnionType, unionName: Name) => void) => void;
    protected forEachUniqueUnion<T>(blankLocations: BlankLineLocations, uniqueValue: (u: UnionType) => T, f: (firstUnion: UnionType, value: T) => void): void;
    protected forEachNamedType: (blankLocations: BlankLineLocations, classFunc: (c: ClassType, className: Name) => void, enumFunc: (e: EnumType, enumName: Name) => void, unionFunc: (u: UnionType, unionName: Name) => void) => void;
    protected sourcelikeToString: (src: Sourcelike) => string;
    protected emitCommentLines: (commentStart: string, lines: string[]) => void;
    private processGraph();
    protected emitSource(givenOutputFilename: string): void;
    protected makeHandlebarsContextForUnionMember(t: Type, name: Name): StringMap;
    protected makeHandlebarsContextForType(t: Type): StringMap;
    protected makeHandlebarsContext(): StringMap;
    protected registerHandlebarsHelpers(context: StringMap): void;
}
